#! /usr/bin/env python3

import pysat
import collections, bz2
import multiprocessing
import datetime, time
#max_years = 1  # Ref: https://github.com/python/cpython/blob/889b0b9bf95651fc05ad532cc4a66c0f8ff29fc2/Include/cpython/pytime.h
#max_sleep_time = datetime.timedelta(days=1).total_seconds()

class QCN:
    """Qualitative Constraint Network (restricted to arcs i < j)"""
    def __init__(self, signature):
        self.size = 0
        self.signature = frozenset(signature)
        s_sign = list(signature)
        s_sign.sort()
        self.universal = s_sign
        self.constraints = dict()

    def add(self, i, j, relation):
        """"Add a constraint (i < j)"""
        self.size = max(self.size, max(i, j)+1)
        assert i < j
        for base_rel in relation:
            if not base_rel in self.signature:
                raise SystemExit("Base relation not in given signature\n")

        idx = (i, j)
        old = self.signature
        if idx in self.constraints:
            old = frozenset(self.constraints[idx])
        relation_sort = list(frozenset(relation) & old)
        relation_sort.sort()

        self.constraints[idx] = relation_sort

    def get(self, i, j):
        """Read constraint on (i,j); always returns a relation."""
        try:
            return self.constraints[(i, j)]
        except KeyError:
            return self.universal

    def is_2_consistent(self):
        """Test 2-consistency"""
        for i, j in self.iterate_strict_triangle():
            if not self.get(i, j):
                return False
        return True

    def iterate_strict_triangle(self):
        """Iterate all pairs i < j on current primal constraint graph."""
        for i in range(0, self.size):
            for j in range(i+1, self.size):
                yield i, j


    def iterate_strict_triples(self):
        """iterate i < j < k triples"""
        for i in range(0, self.size):
            for j in range(i+1, self.size):
                for k in range(j+1, self.size):
                    yield i, j, k


class PropositionalAtoms: 
    """Dictionary enumerating variable strings."""
    def __init__(self):
        self.names = dict()
        self.aux = dict()
        self.r = dict()
        self.last_used = 0
    
    def encode(self, i, j, baserel):
        try:
            return self.names[i][j][baserel]
        except KeyError:
            assert i <= j
            self.last_used += 1

            if not i in self.names:
                self.names[i] = dict()
            if not j in self.names[i]:
                self.names[i][j] = dict()
            self.names[i][j][baserel] = self.last_used
            return self.last_used
            
    def encode_r(self, i, j):
        try:
            return self.r[i][j]
        except KeyError:
            assert i <= j
            self.last_used += 1

            if not i in self.r:
                self.r[i] = dict()
            self.r[i][j] = self.last_used
            return self.last_used
            
    def encode_aux(self, i, j, k, br1, br2, br3):
        try:
            return self.aux[i][j][k][br1][br2][br3]
        except KeyError:
            assert i <= j
            assert j <= k
            self.last_used += 1

            if not i in self.aux:
                self.aux[i] = dict()
            if not j in self.aux[i]:
                self.aux[i][j] = dict()
            if not k in self.aux[i][j]:
                self.aux[i][j][k] = dict()
            if not br1 in self.aux[i][j][k]:
                self.aux[i][j][k][br1] = dict()
            if not br2 in self.aux[i][j][k][br1]:
                self.aux[i][j][k][br1][br2] = dict()
            self.aux[i][j][k][br1][br2][br3] = self.last_used
            return self.last_used

def read_comp_table(calculus):
    """Read an GQR-format composition table"""
    import re

    table = { }
    all_relations = set()

    lines = open(calculus, 'r')
    for line in lines:
        extract = re.search('^(.*):(.*)::[ ]*\((.*)\)$', line)
        assert(extract)
        left = extract.group(1).strip()
        right = extract.group(2).strip()
        supp = extract.group(3).strip()
        table[left + " " + right] = frozenset(supp.split())
        all_relations.add(left)
        all_relations.add(right)
    lines.close()

    return table, frozenset(all_relations)

def read_gqr_csp(buffer, signature):
    """Read an GQR-format CSP from input"""
    import re

    data = QCN(signature)
    
    ID = buffer.pop(0).split()[1].strip()

    for line in buffer:
        res = re.search('^[ ]*([0-9]*)[ ]*([0-9]*) \((.*)\)', line)
        if res:
            i = int(res.group(1))
            j = int(res.group(2))
            assert 0 <= i < j
            rel = res.group(3).strip().split(" ")
            if not rel:
                return None
            for base_rel in rel:
                if not base_rel in signature:
                    raise SystemExit("Calculus signature does not match CSP")

            if frozenset(rel) != signature:  # ignore universal statements
                data.add(i, j, rel)

    return data, ID


def dom_encoding(qcn, atoms, clauses, wclauses):
    for i, j in qcn.iterate_strict_triangle():

        rel = qcn.get(i, j)
        alo = [ atoms.encode(i, j, br) for br in rel ]
        compl_rel = sorted((qcn.signature - set(rel)))
        compl = [ atoms.encode(i, j, br) for br in compl_rel ]

        clauses.append(alo)
        
        for b1 in alo:
           for b2 in [*alo, *compl]: #alo + compl:
              if b1 != b2:
                 clauses.append([-1*b1, -1*b2, atoms.encode_r(i,j)])
                 
        for b1 in [*alo, *compl]:
           wclauses.append(([-1*b1, -1*atoms.encode_r(i,j)],1))
           
                
    
def encoding(qcn, comptable):
    clauses = []
    wclauses = []
    
    atoms = PropositionalAtoms()

    dom_encoding(qcn, atoms, clauses, wclauses)
    
    #from pprint import pprint
    #pprint(atoms.names)

    for i, j, k in qcn.iterate_strict_triples():
        r_ij = qcn.universal
        r_jk = qcn.universal
        r_ik = qcn.universal
        clause = []
        for br1 in r_ij:

            for br2 in r_jk:
                intersection = list(comptable[br1 +" "+ br2])
                intersection.sort()

                for br3 in intersection:               
                    clause += [atoms.encode_aux(i, j, k, br1, br2, br3)]

        clauses.append(clause)   
    

    for i, j, k in qcn.iterate_strict_triples():
        r_ij = qcn.universal
        r_jk = qcn.universal
        r_ik = qcn.universal
        for br1 in r_ij:
            b_ij = 1 * atoms.encode(i, j, br1)
            
            for br2 in r_jk:
                intersection = list(comptable[br1 +" "+ br2])
                intersection.sort()

                b_jk = 1 * atoms.encode(j, k, br2)
                
                for br3 in intersection:  
                    b_ik = 1 * atoms.encode(i, k, br3)       
                    not_aux_var = -1 * atoms.encode_aux(i, j, k, br1, br2, br3)      
                    clauses.append([b_ij, not_aux_var])
                    clauses.append([b_jk, not_aux_var])
                    clauses.append([b_ik, not_aux_var])

    return atoms, clauses, wclauses

def check_options():
    """Parse options with assertions"""
    from argparse import ArgumentParser

    parser = ArgumentParser()

    parser.add_argument('GQR_COMPOSITION_TABLE_FILE')
    parser.add_argument("-f", "--file", action="store", required=True, type=str, dest="filename")
    parser.add_argument("-t", "--timeout", action="store", required=False, type=int, dest="timeout")

    result = parser.parse_args()

    return result.filename, result.timeout,\
        result.GQR_COMPOSITION_TABLE_FILE
        
def foo(wcnf, ID, atoms):
    with RC2S(wcnf) as rc2s:
       model = rc2s.compute()
       if model:
          pararelations = set()
          for i in atoms.r:
             for j in atoms.r[i]:
                 pararelations.add(atoms.r[i][j])
          truemodel = [i for i in model if i>0]
          print("QCN:", ID, rc2s.cost - len([i for i in truemodel if i in pararelations]), rc2s.oracle_time())
          #print("Model:", [i for i in model if i>0])
       else:
          print("QCN:", ID, "no solution")   
       rc2s.delete()           

if __name__ == '__main__':
    QCN_FILENAME, TIMEOUT, CT_FILENAME = check_options()

    COMPTABLE, SIGNATURE = read_comp_table(CT_FILENAME)
    
    from pysat.examples.rc2 import RC2Stratified as RC2S
    from pysat.formula import WCNF
    
    with open(QCN_FILENAME) as f:
       
       total = 0
       meter = 0
       buffer = []

       for i in f:
          buffer.append(i)

          if i.strip() == '.':
             INPUT_QCN, ID = read_gqr_csp(buffer, SIGNATURE)
             buffer.clear()

             if INPUT_QCN.size == 0 or not INPUT_QCN.is_2_consistent():
                # no constraints read (assume problem was unsatisfiable) or
                # problem is not 2-consistency (=> unsat)
                print("p cnf 1 2")
                print("1 0")
                print("-1 0")
                raise SystemExit()

             atoms, clauses, wclauses = encoding(INPUT_QCN, COMPTABLE)

             wcnf = WCNF()            

             for clause in clauses:
                wcnf.append(clause)
       
             for wclause in wclauses:
                wcnf.append(wclause[0], weight=wclause[1])
       
             p = multiprocessing.Process(target=foo, name="Foo", args=(wcnf, ID, atoms))
             p.start()
             p.join(TIMEOUT)
             
             if p.is_alive():
                print("QCN:", ID, "timeout") 

                # Terminate foo
                p.terminate()
                p.join()
